package hr.fer.zemris.diplomskiseminar.mmo.algorithms;

import hr.fer.zemris.diplomskiseminar.mmo.datamodel.DecimalArraySolution;
import hr.fer.zemris.diplomskiseminar.mmo.datamodel.Front;
import hr.fer.zemris.diplomskiseminar.mmo.datamodel.Parameters;
import hr.fer.zemris.diplomskiseminar.mmo.datamodel.Population;
import hr.fer.zemris.diplomskiseminar.mmo.problems.IMMOProblem;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public class NSGA extends GenericAlgorithm{

	private IMMOProblem mmoProblem;
	private int MAX_GEN;
	private double sigmaShare;
	private double epsilon;
	private int POP_SIZE;
	private double CROSS_PROB;
	private double MUT_PROB;
	private int M;
	private int n;
	private double mins;
	private double maxs;
	public final double lamda = 0.2;
	private double minFitness;
	
	public NSGA(IMMOProblem mmoProblem, int maxGen, double sigmaShare,
			double epsilon, int POP_SIZE, double CROSS_PROB, double MUT_PROB, int M) {
		super();
		this.mmoProblem = mmoProblem;
		MAX_GEN = maxGen;
		this.sigmaShare = sigmaShare;
		this.epsilon = epsilon;
		this.POP_SIZE = POP_SIZE;
		this.CROSS_PROB = CROSS_PROB;
		this.MUT_PROB = MUT_PROB;
		this.M = M;
		n = mmoProblem.getDimensionNumber();
		mins = mmoProblem.getMinBound();
		maxs = mmoProblem.getMaxBound();  
	}
	
	public List<Front> start(){
		
		Population population = new Population(POP_SIZE, M, mins, maxs, n);		
		population.initializePopulation();		
		int generationNum = 0;
		List<Front> fronts = null;
		while(generationNum < MAX_GEN){
			System.out.println("Generacija " + generationNum);
			population.evaluatePopulation(mmoProblem);
			fronts = nondominatedSorting(population);
			if(minFitness < 0){
				scaleFitness(population, -minFitness);
			}
			population = reproduction(population);			
			generationNum++;
		}
				
		return fronts;
	}
	
	private void scaleFitness(Population population, double addFitness){
		
		for(DecimalArraySolution s : population.getSolutions()){
			s.TOTAL_FITNESS += addFitness;
		}
	}
	
	
	private Population reproduction(Population old){
		
		Population newPop = new Population(POP_SIZE, M, mins, maxs, n);
		double sum = 0;
		for(DecimalArraySolution solution : old.getSolutions()){
			sum += solution.TOTAL_FITNESS;
		}
		double probability;
		double sumOfProbabilities = 0;
		for(DecimalArraySolution solution : old.getSolutions()){
			probability = solution.TOTAL_FITNESS/sum;
			solution.lowProbability = sumOfProbabilities;
			solution.highProbability = probability + sumOfProbabilities;
			sumOfProbabilities += probability;
		}
		int newSize = 0;
		DecimalArraySolution parent1;
		DecimalArraySolution parent2;
		DecimalArraySolution child1;
		DecimalArraySolution child2;
		while(newSize < POP_SIZE){
			
			parent1 = rouletteWheelSelection(old);
			parent2 = rouletteWheelSelection(old);
			if(Parameters.rand.nextDouble() <= CROSS_PROB){
				child1 = new DecimalArraySolution(M, mins, maxs, n);
				child2 = new DecimalArraySolution(M, mins, maxs, n);
				crossover(parent1, parent2, child1, child2, Parameters.rand, mins, maxs, n);
			}
			else{
				child1 = parent1.duplicate();
				child2 = parent2.duplicate();
			}
			mutate(child1, Parameters.rand, MUT_PROB,mins, maxs, n);
			mutate(child2, Parameters.rand, MUT_PROB, mins, maxs, n);
			newPop.getSolutions().add(child1);
			newPop.getSolutions().add(child2);
			newSize += 2;		
		}
		
		return newPop;
	}
	

	
	private DecimalArraySolution rouletteWheelSelection(Population old){
		double number = Parameters.rand.nextDouble();
		
		for(DecimalArraySolution solution : old.getSolutions()){
			if(number >= solution.lowProbability && number < solution.highProbability){
				return solution;
			}
		}
		throw new RuntimeException("Nitko nije odabran u ruletu!");
	}
	
	public List<Front> nondominatedSorting(Population population){
		
		List<Front> fronts = new ArrayList<>();
		Front f0 = new Front();
		
		for(DecimalArraySolution p : population.getSolutions()){
			p.ni = 0;
			p.setSi(new ArrayList<DecimalArraySolution>());
			for(DecimalArraySolution q : population.getSolutions()){
				int dominates = p.dominates(q);
				if(dominates > 0){
					p.getSi().add(q);
				}
				else if(dominates < 0){
					p.ni++;
				}
			}
			if(p.ni == 0){
				p.rank = 0;
				f0.add(p);
				population.solutionsWithFront++;
			}
		}
		double maxFitnessNextGen = f0.fitnessSharing(POP_SIZE, sigmaShare, n);
		minFitness = maxFitnessNextGen;
		maxFitnessNextGen = maxFitnessNextGen - epsilon;
		fronts.add(f0);
		
		while(!population.isClassified()){
			int frontsNum = fronts.size();
			Front nextFront = new Front();
			for(DecimalArraySolution p : fronts.get(frontsNum-1).solutions){
				Iterator<DecimalArraySolution> iterator = p.getSi().iterator();
				while(iterator.hasNext()){
					DecimalArraySolution q = iterator.next();
					q.ni--;
					if(q.ni == 0){
						q.rank = frontsNum;
						nextFront.add(q);
						population.solutionsWithFront++;
						iterator.remove();
					}
				}
			}		
			if(nextFront.solutions.size()>0){
				maxFitnessNextGen = nextFront.fitnessSharing(maxFitnessNextGen, sigmaShare, n);
				if(maxFitnessNextGen < minFitness){
					minFitness = maxFitnessNextGen;
				}
				maxFitnessNextGen = maxFitnessNextGen - epsilon;
				fronts.add(nextFront);
			}
			else{
				throw new RuntimeException("Logika greka!");
			}
		}
		
		return fronts;	
	}
}










